#define BPF_NO_GLOBAL_DATA
#define __TARGET_ARCH_x86
#include <linux/bpf.h>
#include <bpf/bpf_helpers.h>
#include <bpf/bpf_tracing.h>
#include <bpf/bpf_core_read.h>
#include <linux/if_ether.h>
#include <linux/udp.h>
#include <linux/net.h>
#include <linux/socket.h>
#include <sys/socket.h>
#include <linux/net.h>
#include <arpa/inet.h>
#include <netinet/ip.h>
#include "common.h"


/*
 * Helper:
 * Issue: invalid indirect read from stack R2 off
 *   Fix: check if all variables is initialised
 * Issue: R1 invalid mem access 'inv'
 *   Fix: the value can be NULL
 */

struct {
    __uint(type, BPF_MAP_TYPE_RINGBUF);
    __uint(max_entries, 256 * 1024 /* 256kb */);
} m_data SEC(".maps");

/*
 * This function get the query field and the return the length of it
 */
static size_t get_query(struct __sk_buff *skb, struct event *s_event, uint16_t *class, uint16_t *type, size_t tlen){
    size_t len;
    char buf[QNAME_SIZE] = {0};
    int index = 0;
    int qname_len = 0; // Full length of the qname field
    char qname[QNAME_SIZE] = {0};
    char *c;

    
    bpf_skb_load_bytes(skb, tlen, &buf, 41);
    c = buf;

    /*
     * The qname is composed by a the number of bytes then follow by the label
     * For instance, for the qname www.bucchino.org,
     * the first byte is the number of byte, here, it's 3, then we have www (in hex)
     * Then, we have the byte of 8 and follow by the label bucchino (size 8)
     * And to finish, we have 3 follow by org and we finish with the \0 character
     * For instance, the result is:
     * 03 77 77 77 08 62 75 63 63 68 69 6e 6f 03 6f 72 67 00
     */
    while (*(c++) != '\0') {
        if(*c >= 'a' && *c <= 'z')
            s_event->qname[index] = *c;
        else if(*c >= 'A' && *c <= 'Z')
            s_event->qname[index] = *c;
        else
            s_event->qname[index] = '.';
        index++;
        qname_len++;
    }
    s_event->qname[--index] = '\0';
    qname_len++; // For the null character
    bpf_printk("%s (%d) %d", s_event->qname, index, qname_len);

    // Get class and type
    len = qname_len;
    bpf_skb_load_bytes(skb, tlen + qname_len, type, sizeof(uint16_t));
    len += 2;
    bpf_skb_load_bytes(skb, tlen + qname_len + 2, class, sizeof(uint16_t));
    len += 2;

    return len;
}

/*
 * https://datatracker.ietf.org/doc/html/rfc1035
 */
static int dnsquery(struct __sk_buff *skb, struct ethhdr eth, struct iphdr ip, struct udphdr udp, int dport, int sport){
    struct event *s_event;
    struct dnshdr dns = {0};
    char saddr[32];
    uint16_t class, type;
    // bpf_printk("udp len: %d", ntohs(udp.len));

    s_event = bpf_ringbuf_reserve(&m_data, sizeof(*s_event), 0);
    if (!s_event)
       return 0;

    /* Get IP header */
    s_event->saddr = ip.saddr;

    /* Get DNS header */
    bpf_skb_load_bytes(skb, sizeof(struct ethhdr) + sizeof(struct iphdr) + sizeof(struct udphdr), &dns, sizeof(struct dnshdr));

    if (ntohs(dns.nbQuestions) == 0){
        bpf_ringbuf_discard(s_event, 0);
        return 0;
    }


    bpf_printk("tid: %x", ntohs(dns.transactionID)); // Use as key map
    bpf_printk("nb question: %d", ntohs(dns.nbQuestions));

    //struct dns_query dquery;
    //bpf_skb_load_bytes(skb, sizeof(struct ethhdr) + sizeof(struct iphdr) + sizeof(struct udphdr) + sizeof(struct dnshdr), &dquery, sizeof(struct dns_query));
    // bpf_printk("size: %d %d %d", tlen, skb->len, (skb->len - tlen));
    //dlen = (skb->len - tlen);
    //bpf_printk("DNS packet len: %d", dlen);
    //qlen = dlen - sizeof(struct dnshdr);
    //bpf_printk("size: %d %d", sizeof(struct dnshdr), qlen);


    /* Get the query structure */
    size_t tlen = sizeof(struct ethhdr) + sizeof(struct iphdr) + sizeof(struct udphdr) + sizeof(struct dnshdr);
    size_t query_len = get_query(skb, s_event, &class, &type, tlen);

    // https://docs.cilium.io/en/stable/reference-guides/bpf/progtypes/
    s_event->dport = dport;
    s_event->sport = sport;
    s_event->class = ntohs(class);
    s_event->type = ntohs(type);
    //if(bpf_probe_read_user_str(&s_event->qname, sizeof(s_event->qname), qname) < 0)
    //    bpf_printk("Failed to copy qname");

    // Add to map

    bpf_ringbuf_submit(s_event, 0);

    return 0;
}

static int dnsanswer(struct __sk_buff *skb, struct ethhdr eth, struct iphdr ip, struct udphdr udp, int dport, int sport){
    return 0;
}
SEC("socket")
int detect_dns(struct __sk_buff *skb) {
    //void *data = (void *)(long)skb->data;
    //void *data_end = (void *)(long)skb->data_end;
    //struct ethhdr *eth = data;
    struct ethhdr eth = {0};
    struct iphdr ip = {0};
    struct udphdr udp = {0};
    unsigned long long h_proto, p;
    unsigned long long dport;
    unsigned long long sport;

    //if (data + sizeof(struct ethhdr) + sizeof(struct iphdr) + sizeof(struct udphdr) > data_end)
    //    return 0;

    //bpf_skb_load_bytes(skb, 12, &p, 2);
    bpf_skb_load_bytes(skb, 0, &eth, sizeof(struct ethhdr));
    p = eth.h_proto;
    if (ntohs(p) != ETH_P_IP)
        return 0;

    // bpf_printk("ip: %d",ntohs(p));

    //ip = (struct iphdr*)(data + sizeof(struct ethhdr));
    bpf_skb_load_bytes(skb, sizeof(struct ethhdr), &ip, sizeof(struct iphdr));

    //bpf_skb_load_bytes(data, sizeof(struct ethhdr), ip, sizeof(struct ip));
    h_proto = ip.protocol;
    // If not UDP packet
    if (h_proto != 17) 
        return 0;

    // bpf_printk("proto: %d", h_proto);
    //udp = (struct udphdr*)(data + sizeof(struct ethhdr) + sizeof(struct iphdr));

    bpf_skb_load_bytes(skb, sizeof(struct ethhdr) + sizeof(struct iphdr), &udp, sizeof(struct udphdr));

    // Check if DNS port
    dport = ntohs(udp.dest);
    sport = ntohs(udp.source);
    
    if (dport == 53)
        dnsquery(skb, eth, ip, udp, dport, sport);
    else if(sport == 53)
        bpf_printk("Response");

    return 0;
}
/*SEC("xdp")
int detect_dns(struct xdp_md *ctx){
    void *data_end = (void *)(long)ctx->data_end;
    void *data = (void *)(long)ctx->data;
    struct ethhdr *eth = data;
    struct iphdr *ip;
    struct udphdr *udp;
    __u16 h_proto;
    __u16 dport;

    if (data + sizeof(struct ethhdr) + sizeof(struct iphdr) + sizeof(struct udphdr) > data_end)
        return XDP_DROP;

    ip = (struct iphdr*)(data + sizeof(struct ethhdr));
    udp = (struct udphdr*)(data + sizeof(struct ethhdr) + sizeof(struct iphdr));

    h_proto = ip->protocol;
    // If not UDP packet
    if (h_proto != 17) 
        return XDP_PASS;

    // Check if DNS port
    dport = udp->dest;
    bpf_printk("Dport: %d", (dport));

    return XDP_PASS;
}*/
char LICENSE[] SEC("license") = "GPL";
